#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <inttypes.h>
#include <ctype.h>

#include <xenctrl.h>
#include "xen/xc_dom.h"

#include "msr-index.h"
#include "xenner.h"
#include "mm.h"

/* ------------------------------------------------------------------ */

static const struct kvm_segment xen32_cs0 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe008,
    .dpl      = 0,
    .type     = 0xb,
    .present  = 1,  .db = 1,  .s = 1,  .g = 1,
};
static const struct kvm_segment xen32_ds0 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe010,
    .dpl      = 0,
    .type     = 0x3,
    .present  = 1,  .db = 1,  .s = 1,  .g = 1,
};
static const struct kvm_segment xen32_cs1 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe019,
    .dpl      = 1,
    .type     = 0xb,
    .present  = 1,  .db = 1,  .s = 1,  .g = 1,
};
static const struct kvm_segment xen32_ds1 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe021,
    .dpl      = 1,
    .type     = 0x3,
    .present  = 1,  .db = 1,  .s = 1,  .g = 1,
};
static const struct kvm_segment xen32_cs3 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe02b,
    .dpl      = 3,
    .type     = 0xb,
    .present  = 1,  .db = 1,  .s = 1,  .g = 1,
};
static const struct kvm_segment xen32_ds3 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe033,
    .dpl      = 3,
    .type     = 0x3,
    .present  = 1,  .db = 1,  .s = 1,  .g = 1,
};

static const struct kvm_segment xen64_cs0_64 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe008,
    .dpl      = 0,
    .type     = 0xb,
    .present  = 1,  .l = 1,  .s = 1,  .g = 1,
};
static const struct kvm_segment xen64_ds0_32 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe010,
    .dpl      = 0,
    .type     = 0x3,
    .present  = 1,  .db = 1,  .s = 1,  .g = 1,
};
static const struct kvm_segment xen64_cs3_32 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe023,
    .dpl      = 3,
    .type     = 0xb,
    .present  = 1,  .db = 1,  .s = 1,  .g = 1,
};
static const struct kvm_segment xen64_ds3_32 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe02b,
    .dpl      = 3,
    .type     = 0x3,
    .present  = 1,  .db = 1,  .s = 1,  .g = 1,
};
static const struct kvm_segment xen64_cs3_64 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe033,
    .dpl      = 3,
    .type     = 0xb,
    .present  = 1,  .l = 1,  .s = 1,  .g = 1,
};
static const struct kvm_segment xen64_cs0_32 = {
    .base     = 0,
    .limit    = 0xffffffff,
    .selector = 0xe038,
    .dpl      = 0,
    .type     = 0xb,
    .present  = 1,  .db = 1,  .s = 1,  .g = 1,
};

static const struct kvm_segment noseg = {
    .unusable = 1,
};

static struct descriptor_32 mkdesc(const struct kvm_segment *seg)
{
    struct descriptor_32 desc;
    int shift = 0;

    shift  = seg->g ? 12 : 0;
    desc.a = (seg->limit >> shift) & 0xffff;
    desc.b = (seg->limit >> shift) & 0x000f0000;
    
    desc.a |= (seg->base & 0xffff) << 16;
    desc.b |= seg->base & 0xff000000;
    desc.b |= (seg->base & 0xff0000) >> 16;
    desc.b |= (seg->type & 0x0f) << 8;
    desc.b |= (seg->dpl & 0x03) << 13;

    if (seg->s)       desc.b |= (1 << 12);
    if (seg->present) desc.b |= (1 << 15);
    if (seg->avl)     desc.b |= (1 << 20);
    if (seg->l)       desc.b |= (1 << 21);
    if (seg->db)      desc.b |= (1 << 22);
    if (seg->g)       desc.b |= (1 << 23);

#if 0
    fprintf(stderr,
	    "%s: [%04x] %08llx +%08x type %x dpl %x [ %s %s %s %s %s %s] -> %08x:%08x\n",
	    __FUNCTION__, seg->selector,
	    seg->base, seg->limit, seg->type, seg->dpl,
	    seg->s       ? "s"  : " ",
	    seg->present ? "p"  : " ",
	    seg->avl     ? "a"  : " ",
	    seg->l       ? "l"  : " ",
	    seg->db      ? "db" : "  ",
	    seg->g       ? "g"  : " ",
	    desc.b, desc.a);
#endif

    return desc;
}

static inline void gdt_set(struct xenvcpu *v, const struct kvm_segment *seg)
{
    v->gdt[ seg->selector >> 3 ] = mkdesc(seg);
}

/* ------------------------------------------------------------------ */

static void setup_regs_all(struct xenvm *xen,
			   struct xc_dom_image *dom)
{
    struct xenvcpu *vcpu0 = xen->vcpu + 0;  /* boot cpu */
    pfn_t cr3_pfn;

    /* GP registers */
    vcpu0->regs.rip = elf_uval(xen->emu, xen->emu->ehdr, e_entry);

    /* protected mode */
    vcpu0->sregs.cr0 |= X86_CR0_PE;
    vcpu0->sregs.ldt = noseg;

    /* paging setup */
    cr3_pfn = xen->emu_pgd_mfn;
    vcpu0->sregs.cr3 = frame_to_addr(cr3_pfn);
    vcpu0->sregs.cr0 |= X86_CR0_PG; /* paging */
    vcpu0->sregs.cr0 |= X86_CR0_WP;
    vcpu0->sregs.cr4 |= X86_CR4_PSE; /* large pages */
    vcpu0->sregs.cr4 |= X86_CR4_PGE; /* global pages */
}

static void setup_regs_32(struct xenvm *xen,
			  struct xc_dom_image *dom)
{
    struct xenvcpu *vcpu0 = xen->vcpu + 0;  /* boot cpu */

    /* 32-bit protected mode setup */
    vcpu0->sregs.cs  = xen32_cs0;
    vcpu0->sregs.ds  = xen32_ds0;
    vcpu0->sregs.es  = vcpu0->sregs.ds;
    vcpu0->sregs.fs  = vcpu0->sregs.ds;
    vcpu0->sregs.gs  = vcpu0->sregs.ds;
    vcpu0->sregs.ss  = vcpu0->sregs.ds;

    /* paging */
    if (xen->mode == XENMODE_PAE)
	vcpu0->sregs.cr4 |= X86_CR4_PAE;
}

static void setup_regs_64(struct xenvm *xen,
			  struct xc_dom_image *dom)
{
    struct xenvcpu *vcpu0 = xen->vcpu + 0;  /* boot cpu */

    /* 64-bit long mode setup */
    vcpu0->sregs.cs  = xen64_cs0_64;
    vcpu0->sregs.ds  = xen64_ds0_32;
    vcpu0->sregs.es  = vcpu0->sregs.ds;
    vcpu0->sregs.fs  = vcpu0->sregs.ds;
    vcpu0->sregs.gs  = vcpu0->sregs.ds;
    vcpu0->sregs.ss  = vcpu0->sregs.ds;

    vcpu0->sregs.efer |= EFER_SCE;   // syscall enable
    vcpu0->sregs.efer |= EFER_LME;   // long mode enable
    vcpu0->sregs.efer |= EFER_LMA;   // long mode active

    /* paging */
    vcpu0->sregs.cr4 |= X86_CR4_PAE;
}

static void setup_regs(struct xenvm *xen,
		       struct xc_dom_image *dom)
{
    switch (xen->mode) {
    case XENMODE_32:
    case XENMODE_PAE:
	d1printf("%s: 32bit\n", __FUNCTION__);
	setup_regs_all(xen, dom);
	setup_regs_32(xen, dom);
	break;
    case XENMODE_64:
	d1printf("%s: 64bit\n", __FUNCTION__);
	setup_regs_all(xen, dom);
	setup_regs_64(xen, dom);
	break;
    default:
	break;
    }
}

/* ------------------------------------------------------------------ */

static enum xenmode xen_guest_mode(struct xc_dom_image *dom)
{
    if (0 == strcmp(dom->guest_type, "xen-3.0-x86_32")) {
	return XENMODE_32;
    } else if (0 == strcmp(dom->guest_type, "xen-3.0-x86_32p")) {
	return XENMODE_PAE;
    } else if (0 == strcmp(dom->guest_type, "xen-3.0-x86_64")) {
	return XENMODE_64;
    } else {
	return 0;
    }
}

static void xen_guest_hypercall_page_init_32(void *hypercall_page)
{
    char *p;
    int i;

    /* Fill in all the transfer points with template machine code. */
    for ( i = 0; i < (PAGE_SIZE / 32); i++ )
    {
        p = (char *)(hypercall_page + (i * 32));
        *(uint8_t  *)(p+ 0) = 0xb8;    /* mov  $<i>,%eax */
        *(uint32_t *)(p+ 1) = i;
        *(uint16_t *)(p+ 5) = 0x82cd;  /* int  $0x82 */
        *(uint8_t  *)(p+ 7) = 0xc3;    /* ret */
    }

    /*
     * HYPERVISOR_iret is special because it doesn't return and expects a 
     * special stack frame. Guests jump at this transfer point instead of 
     * calling it.
     */
    p = (char *)(hypercall_page + (__HYPERVISOR_iret * 32));
    *(uint8_t  *)(p+ 0) = 0x50;    /* push %eax */
    *(uint8_t  *)(p+ 1) = 0xb8;    /* mov  $__HYPERVISOR_iret,%eax */
    *(uint32_t *)(p+ 2) = __HYPERVISOR_iret;
    *(uint16_t *)(p+ 6) = 0x82cd;  /* int  $0x82 */
}

static void xen_guest_hypercall_page_init_64(void *hypercall_page)
{
    char *p;
    int i;

    /* Fill in all the transfer points with template machine code. */
    for ( i = 0; i < (PAGE_SIZE / 32); i++ )
    {
        p = (char *)(hypercall_page + (i * 32));
        *(uint8_t  *)(p+ 0) = 0x51;    /* push %rcx */
        *(uint16_t *)(p+ 1) = 0x5341;  /* push %r11 */
        *(uint8_t  *)(p+ 3) = 0xb8;    /* mov  $<i>,%eax */
        *(uint32_t *)(p+ 4) = i;
        *(uint16_t *)(p+ 8) = 0x050f;  /* syscall */
        *(uint16_t *)(p+10) = 0x5b41;  /* pop  %r11 */
        *(uint8_t  *)(p+12) = 0x59;    /* pop  %rcx */
        *(uint8_t  *)(p+13) = 0xc3;    /* ret */
    }

    /*
     * HYPERVISOR_iret is special because it doesn't return and expects a 
     * special stack frame. Guests jump at this transfer point instead of 
     * calling it.
     */
    p = (char *)(hypercall_page + (__HYPERVISOR_iret * 32));
    *(uint8_t  *)(p+ 0) = 0x51;    /* push %rcx */
    *(uint16_t *)(p+ 1) = 0x5341;  /* push %r11 */
    *(uint8_t  *)(p+ 3) = 0x50;    /* push %rax */
    *(uint8_t  *)(p+ 4) = 0xb8;    /* mov  $__HYPERVISOR_iret,%eax */
    *(uint32_t *)(p+ 5) = __HYPERVISOR_iret;
    *(uint16_t *)(p+ 9) = 0x050f;  /* syscall */
}

static void xen_guest_copy(struct xenvm *xen, struct xc_dom_image *dom)
{
    struct xc_dom_phys *phys;
    void *dest;

    for (phys = dom->phys_pages; phys != NULL; phys = phys->next) {
	dest = mfn_to_ptr(xen, xen->mfn_guest + phys->first);
	d1printf("%s: 0x%04" PRIpfn " +0x%04" PRIpfn " @ %p => +%lx\n",
		 __FUNCTION__, phys->first, phys->count, phys->ptr,
		 (unsigned long)(dest - xen->memory));
	memcpy(dest, phys->ptr, phys->count * PAGE_SIZE);
    }
}

static int xen_guest_parse(struct xenvm *xen, struct xc_dom_image *dom)
{
    int rc;

    section_print(xen, __FUNCTION__, "parse kernel");
    if (0 != (rc = xc_dom_kernel_file(dom, xen->kernel)))
	goto out;
    if (xen->ramdisk && strlen(xen->ramdisk))
	if (0 != (rc = xc_dom_ramdisk_file(dom, xen->ramdisk)))
	    goto out;
    if (0 != (rc = xc_dom_parse_image(dom)))
	goto out;
    return 0;

out:
    section_print(xen, __FUNCTION__, "FAILURE");
    return -1;
}

static int xen_guest_setup(struct xenvm *xen, struct xc_dom_image *dom)
{
    void *ptr;
    xen_pfn_t pfn;
    int rc;

    section_print(xen, __FUNCTION__, "memory setup");
    if (0 != (rc = xc_dom_mem_init(dom, PG_TO_MB(xen->pg_guest))))
	goto out;
    dom->p2m_host = xc_dom_malloc(dom, sizeof(xen_pfn_t) * dom->total_pages);
    for (pfn = 0; pfn < dom->total_pages; pfn++)
        dom->p2m_host[pfn] = pfn + xen->mfn_guest;

    section_print(xen, __FUNCTION__, "create start-of-day");
    if (0 != (rc = xc_dom_build_image(dom)))
	goto out;
    if (0 != (rc = xc_dom_update_guest_p2m(dom)))
	goto out;
    if (0 != (rc = dom->arch_hooks->setup_pgtables(dom)))
	goto out;

    ptr = xc_dom_pfn_to_ptr(dom, dom->console_pfn, 1);
    memset(ptr, 0, PAGE_SIZE);
    ptr = xc_dom_pfn_to_ptr(dom, dom->xenstore_pfn, 1);
    memset(ptr, 0, PAGE_SIZE);

    dom->arch_hooks->start_info(dom);

    if (dom->parms.virt_hypercall) {
	section_print(xen, __FUNCTION__, "setup xen hypercall page");
	pfn = addr_to_frame(dom->parms.virt_hypercall - dom->parms.virt_base);
	ptr = xc_dom_pfn_to_ptr(dom, pfn, 1);
	switch (xen->mode) {
	case XENMODE_32:
	case XENMODE_PAE:
	    xen_guest_hypercall_page_init_32(ptr);
	    break;
	case XENMODE_64:
	    xen_guest_hypercall_page_init_64(ptr);
	    break;
	}
    } else {
	section_print(xen, __FUNCTION__, "warn: no hypercall page");
    }

    section_print(xen, __FUNCTION__, "setup vcpu context");
    if (0 != (rc = dom->arch_hooks->vcpu(dom, xen->vcpu[0].ct.ptr)))
	goto out;

    return 0;

out:
    section_print(xen, __FUNCTION__, "FAILURE");
    return -1;
}

/* ------------------------------------------------------------------ */

static const char emu32_blob[] = {
#include "emu32-blob.h"
};
#ifdef __x86_64__
static const char emu64_blob[] = {
#include "emu64-blob.h"
};
#endif

static int xen_emu_load(struct xenvm *xen, struct xc_dom_image *dom)
{
    section_print(xen, __FUNCTION__, "load xen emu");
    xen->emu = malloc(sizeof(struct elf_binary));
    switch (xen->mode) {
    case XENMODE_32:
    case XENMODE_PAE:
	elf_init(xen->emu, emu32_blob, sizeof(emu32_blob));
	break;
    case XENMODE_64:
#ifdef __x86_64__
	elf_init(xen->emu, emu64_blob, sizeof(emu64_blob));
#else
	return -1;
#endif
	break;
    }
    if (xen->logfile)
	elf_set_logfile(xen->emu, xen->logfile, 1);
    else if (xen->debug && !xen->nostderr)
	elf_set_logfile(xen->emu, stderr, 1);
    elf_parse_binary(xen->emu);
    xen->emu->dest = mfn_to_ptr(xen, xen->mfn_emu);
    elf_load_binary(xen->emu);
    xen->emu_vs = xen->emu->pstart;
    xen->emu_ve = xen->emu->pend;
    return 0;
}

static int xen_emu32_setup(struct xenvm *xen, struct xc_dom_image *dom)
{
    struct xenvcpu *vcpu0 = xen->vcpu + 0;  /* boot cpu */
    struct vcpu_time_info_x86_32 *time;
    uint64_t addr;

    section_print(xen, __FUNCTION__, "gdt setup");
    addr = elf_lookup_addr(xen->emu, "xen_gdt");
    vcpu0->gdt = emu_vaddr_to_ptr(xen, addr);
    vcpu0->sregs.gdt.base = addr;
    gdt_set(vcpu0, &xen32_cs0);
    gdt_set(vcpu0, &xen32_ds0);
    gdt_set(vcpu0, &xen32_cs1);
    gdt_set(vcpu0, &xen32_ds1);
    gdt_set(vcpu0, &xen32_cs3);
    gdt_set(vcpu0, &xen32_ds3);

    section_print(xen, __FUNCTION__, "xen cpu context setup");
    addr = elf_lookup_addr(xen->emu, "xen_ctxt");
    vcpu0->ct.ct32 = emu_vaddr_to_ptr(xen, addr);

    section_print(xen, __FUNCTION__, "shared_info setup");
    addr = elf_lookup_addr(xen->emu, "shared_info");
    dom->shared_info_mfn = xen->mfn_emu + 
	addr_to_frame(addr - xen->emu_vs);

    xen->shinfo.sh32 = emu_vaddr_to_ptr(xen, addr);
    xen->xen_wc = (void*)(&xen->shinfo.sh32->wc_version);

    vcpu0->evtchn_upcall_pending =
	&(xen->shinfo.sh32->vcpu_info[0].evtchn_upcall_pending);
    vcpu0->evtchn_upcall_mask =
	&(xen->shinfo.sh32->vcpu_info[0].evtchn_upcall_mask);
    time = &(xen->shinfo.sh32->vcpu_info[0].time);
    vcpu0->xentime = (struct vcpu_time_info*)time;

    section_print(xen, __FUNCTION__, "emu pgd setup");
    if (XENMODE_PAE == xen->mode)
	addr = elf_lookup_addr(xen->emu, "emu_pgd_pae");
    else
	addr = elf_lookup_addr(xen->emu, "emu_pgd_32");
    xen->emu_pgd_mfn = xen->mfn_emu + addr_to_frame(addr - xen->emu_vs);
    
    return 0;
}

static int xen_emu64_setup(struct xenvm *xen, struct xc_dom_image *dom)
{
    struct xenvcpu *vcpu0 = xen->vcpu + 0;  /* boot cpu */
    uint64_t addr;
    struct vcpu_time_info_x86_64 *time;

    section_print(xen, __FUNCTION__, "gdt setup");
    addr = elf_lookup_addr(xen->emu, "xen_gdt");
    vcpu0->gdt = emu_vaddr_to_ptr(xen, addr);
    vcpu0->sregs.gdt.base = addr;
    gdt_set(vcpu0, &xen64_cs0_64);
    gdt_set(vcpu0, &xen64_ds0_32);
    gdt_set(vcpu0, &xen64_cs3_32);
    gdt_set(vcpu0, &xen64_ds3_32);
    gdt_set(vcpu0, &xen64_cs3_64);
    gdt_set(vcpu0, &xen64_cs0_32);

    section_print(xen, __FUNCTION__, "xen cpu context setup");
    addr = elf_lookup_addr(xen->emu, "xen_ctxt");
    vcpu0->ct.ct64 = emu_vaddr_to_ptr(xen, addr);
    
    section_print(xen, __FUNCTION__, "shared_info setup");
    addr = elf_lookup_addr(xen->emu, "shared_info");
    dom->shared_info_mfn = xen->mfn_emu + 
	addr_to_frame(addr - xen->emu_vs);

    xen->shinfo.sh64 = emu_vaddr_to_ptr(xen, addr);
    xen->xen_wc = (void*)(&xen->shinfo.sh64->wc_version);

    vcpu0->evtchn_upcall_pending =
	&(xen->shinfo.sh64->vcpu_info[0].evtchn_upcall_pending);
    vcpu0->evtchn_upcall_mask =
	&(xen->shinfo.sh64->vcpu_info[0].evtchn_upcall_mask);
    time = &(xen->shinfo.sh64->vcpu_info[0].time);
    vcpu0->xentime = (struct vcpu_time_info*)time;

    section_print(xen, __FUNCTION__, "emu pgd setup");
    addr = elf_lookup_addr(xen->emu, "emu_pgd");
    xen->emu_pgd_mfn = xen->mfn_emu + addr_to_frame(addr - xen->emu_vs);
    
    return 0;
}

static int xen_emu_setup(struct xenvm *xen, struct xc_dom_image *dom)
{
    uint32_t *ptr32;
    uint64_t *ptr64;
    uint64_t addr;

    /* common stuff */
    section_print(xen, __FUNCTION__, "stats setup");
    addr = elf_lookup_addr(xen->emu, "emu_hcalls");
    xen->emu_hcalls = emu_vaddr_to_ptr(xen, addr);
    addr = elf_lookup_addr(xen->emu, "emu_faults");
    xen->emu_faults = emu_vaddr_to_ptr(xen, addr);
    addr = elf_lookup_addr(xen->emu, "grant_table");
    xen->grant_table = emu_vaddr_to_ptr(xen, addr);

    section_print(xen, __FUNCTION__, "memory info");
    addr = elf_lookup_addr(xen->emu, "mfn_emu");
    ptr64 = emu_vaddr_to_ptr(xen, addr);
    *ptr64 = xen->mfn_emu;
    addr = elf_lookup_addr(xen->emu, "pg_emu");
    ptr64 = emu_vaddr_to_ptr(xen, addr);
    *ptr64 = xen->pg_emu;
    addr = elf_lookup_addr(xen->emu, "mfn_m2p");
    ptr64 = emu_vaddr_to_ptr(xen, addr);
    *ptr64 = xen->mfn_m2p;
    addr = elf_lookup_addr(xen->emu, "pg_m2p");
    ptr64 = emu_vaddr_to_ptr(xen, addr);
    *ptr64 = xen->pg_m2p;
    addr = elf_lookup_addr(xen->emu, "mfn_guest");
    ptr64 = emu_vaddr_to_ptr(xen, addr);
    *ptr64 = xen->mfn_guest;
    addr = elf_lookup_addr(xen->emu, "pg_guest");
    ptr64 = emu_vaddr_to_ptr(xen, addr);
    *ptr64 = xen->pg_guest;
    addr = elf_lookup_addr(xen->emu, "pg_total");
    ptr64 = emu_vaddr_to_ptr(xen, addr);
    *ptr64 = xen->pg_total;

    addr = elf_lookup_addr(xen->emu, "debug_level");
    ptr32 = emu_vaddr_to_ptr(xen, addr);
    *ptr32 = xen->debug;

    switch (xen->mode) {
    case XENMODE_32:
    case XENMODE_PAE:
	return xen_emu32_setup(xen, dom);
    case XENMODE_64:
	return xen_emu64_setup(xen, dom);
    }
    return 0;
}

/* ------------------------------------------------------------------ */

int domain_builder(struct xenvm *xen)
{
    struct xc_dom_image *dom = NULL;
    pfn_t pfn;

    if (xen->logfile)
	xc_dom_logfile = xen->logfile;
    else if (xen->debug && !xen->nostderr)
	xc_dom_logfile = stderr;
    dom = xc_dom_allocate(xen->cmdline, 0);

    if (0 != xen_guest_parse(xen, dom))
	goto out;
    xen->mode = xen_guest_mode(dom);
    switch (xen->mode) {
    case XENMODE_32:
    case XENMODE_PAE:
	/* alloc memory */
	xen->mfn_emu   = 0;
	xen->pg_emu    = MB_TO_PG(4);
	xen->mfn_m2p   = xen->pg_emu;
	xen->pg_m2p    = MB_TO_PG(4);
	xen->m2p_32    = mfn_to_ptr(xen, xen->mfn_m2p);
	xen->mfn_guest = xen->pg_emu + xen->pg_m2p;
	xen->pg_guest  = xen->pg_total - xen->mfn_guest;
	break;
    case XENMODE_64:
	xen->mfn_emu   = 0;
	xen->pg_emu    = MB_TO_PG(4);
	xen->mfn_m2p   = xen->pg_emu;
	xen->pg_m2p    = MB_TO_PG(4);
	xen->m2p_64    = mfn_to_ptr(xen, xen->mfn_m2p);
	xen->mfn_guest = xen->pg_emu + xen->pg_m2p;
	xen->pg_guest  = xen->pg_total - xen->mfn_guest;
	break;
    default:
	d0printf("%s: unknown guest mode: %d (%s)\n",  __FUNCTION__,
		 xen->mode, dom->guest_type);
	goto out;
    }

    dom->console_evtchn  = xen->console_event;
    dom->xenstore_evtchn = xen->xenstore_event;

    if (0 != xen_emu_load(xen, dom))
	goto out;
    if (0 != xen_emu_setup(xen, dom))
	goto out;
    if (0 != xen_guest_setup(xen, dom))
	goto out;

    xen->console_pfn = dom->console_pfn;
    xen->xenstore_pfn = dom->xenstore_pfn;

    /* m2p setup */
    if (xen->m2p_32)
	for (pfn = 0; pfn < dom->total_pages; pfn++)
	    xen->m2p_32[pfn + xen->mfn_guest] = pfn;
    if (xen->m2p_64)
	for (pfn = 0; pfn < dom->total_pages; pfn++)
	    xen->m2p_64[pfn + xen->mfn_guest] = pfn;

    section_print(xen, __FUNCTION__, "kvm: state setup");
    xen_guest_copy(xen, dom);
    setup_regs(xen, dom);

    xc_dom_release(dom);
    section_print(xen, __FUNCTION__, "all done");
    return 0;

out:
    section_print(xen, __FUNCTION__, "FAILURE");
    return -1;
}
