#include "emu.h"
#include "emu-main.c"

/* --------------------------------------------------------------------- */
/* inline asm bits                                                       */

int is_pae(void)
{
    static int pae = -1;

    if (-1 == pae) {
	pae = (read_cr4() & X86_CR4_PAE);
	printk(1, "pae is turned %s\n", pae ? "on" : "off");
    }
    return pae;
}

/* --------------------------------------------------------------------- */
/* helper functions                                                      */

int bounce_trap(struct regs_32 *regs, int trapno, int cbno)
{
    uint32_t *kesp, eip, cs;
    int stack_switch = 0;
    int error_code = 0;
    int interrupt = 0;
    int k = 0;

    if (trapno >= 0) {
	/* trap bounce */
	eip  = xen_ctxt.trap_ctxt[trapno].address;
	cs   = xen_ctxt.trap_ctxt[trapno].cs;
	if (TI_GET_IF(&xen_ctxt.trap_ctxt[trapno]))
	    interrupt = 1;
	if (trapno < sizeof(trapinfo)/sizeof(trapinfo[0]))
	    error_code = trapinfo[trapno].ec;
	if (14 == trapno) /* page fault */
	    shared_info.vcpu_info[0].arch.cr2 = read_cr2();
    }
    if (cbno >= 0) {
	/* callback */
	eip  = xencb[cbno].eip;
	cs   = xencb[cbno].cs;
	switch (cbno) {
	case CALLBACKTYPE_event:
 	    interrupt = 1;
	    break;
	}
    }

    if (!cs)
	panic("no guest trap handler");

    if ((regs->cs & 0x03) < (cs & 0x03))
	panic("bounce trap: illegal ring switch\n");
    if ((regs->cs & 0x03) > (cs & 0x03))
	stack_switch = 1;

    /* prepare guest stack: copy from emu, so the handler
     * jumps straigt back without round-trip via emu */
    if (stack_switch) {
	kesp = (void*)xen_tss.esp1;
	kesp[-(++k)] = regs->ss;            // push ss
	kesp[-(++k)] = regs->esp;           // push esp
    } else {
	kesp = (void*)regs->esp;
    }

    kesp[-(++k)] = regs->eflags;            // push eflags
    kesp[-(++k)] = regs->cs;                // push cs
    kesp[-(++k)] = regs->eip;               // push eip
    if (error_code)
	kesp[-(++k)] = regs->error;         // push error code

    /* prepare emu stack, so iret jumps to the kernels handler. */
    regs->eip     = eip;
    regs->cs      = cs;
    regs->eflags &= EFLAGS_TRAPMASK;
    if (stack_switch) {
	regs->ss  = xen_tss.ss1;
	regs->esp = xen_tss.esp1;
    }
    regs->esp -= 4*k;

    return 0;
}

static void tss_init(int vcpu)
{
    int idx = tss(vcpu);

    xen_tss.esp0 = (intptr_t)stack_high;
    xen_tss.ss0  = 0xe010;

    xen_gdt[ idx ] = mkdesc32((intptr_t)(&xen_tss), sizeof(xen_tss)-1, 0x89, 0);
    ltr(idx << 3);
}

static void idt_init(void)
{
    struct {
	uint16_t len;
	uint32_t ptr;
    } __attribute__((packed)) idtp = {
	.len = sizeof(xen_idt)-1,
	.ptr = (intptr_t)xen_idt,
    };
    int i;

    for (i = 0; i < 256; i++)
	xen_idt[i] = mkgate32(0xe008, (intptr_t)int_unknown, 0x8e);

    xen_idt[    0 ] = mkgate32(0xe008, (intptr_t)division_by_zero,    0x8e);
    xen_idt[    1 ] = mkgate32(0xe008, (intptr_t)debug_int1,          0x8e);
    xen_idt[    2 ] = mkgate32(0xe008, (intptr_t)nmi,                 0x8e);
    xen_idt[    3 ] = mkgate32(0xe008, (intptr_t)debug_int3,          0xee);
    xen_idt[    4 ] = mkgate32(0xe008, (intptr_t)overflow,            0x8e);
    xen_idt[    5 ] = mkgate32(0xe008, (intptr_t)bound_check,         0x8e);
    xen_idt[    6 ] = mkgate32(0xe008, (intptr_t)illegal_instruction, 0x8e);
    xen_idt[    7 ] = mkgate32(0xe008, (intptr_t)no_device,           0x8e);
    xen_idt[    8 ] = mkgate32(0xe008, (intptr_t)double_fault,        0x8e);
    xen_idt[    9 ] = mkgate32(0xe008, (intptr_t)coprocessor,         0x8e);
    xen_idt[   10 ] = mkgate32(0xe008, (intptr_t)invalid_tss,         0x8e);
    xen_idt[   11 ] = mkgate32(0xe008, (intptr_t)segment_not_present, 0x8e);
    xen_idt[   12 ] = mkgate32(0xe008, (intptr_t)stack_fault,         0x8e);
    xen_idt[   13 ] = mkgate32(0xe008, (intptr_t)general_protection,  0x8e);
    xen_idt[   14 ] = mkgate32(0xe008, (intptr_t)page_fault,          0x8e);
    xen_idt[   16 ] = mkgate32(0xe008, (intptr_t)floating_point,      0x8e);
    xen_idt[   17 ] = mkgate32(0xe008, (intptr_t)alignment,           0x8e);
    xen_idt[   18 ] = mkgate32(0xe008, (intptr_t)machine_check,       0x8e);
    xen_idt[   19 ] = mkgate32(0xe008, (intptr_t)simd_floating_point, 0x8e);

    xen_idt[ 0x20 ] = mkgate32(0xe008, (intptr_t)event_callback,      0x8e);
    xen_idt[ 0x82 ] = mkgate32(0xe008, (intptr_t)xen_hypercall,       0xae);
    
    asm volatile("lidt %0" : : "m" (idtp));
}

/* --------------------------------------------------------------------- */

static int pf_fixup_readonly(struct regs_32 *regs, uint32_t cr2)
{
    if (!is_pae()) {
	uint32_t *pte = find_pte_32_lpt(cr2);
	if (cr2 >= XEN_M2P_32)
	    return 0;
	if (*pte & _PAGE_USER)
	    return 0;
	/* is kernel page */
	*pte |= _PAGE_RW;
    } else {
	uint64_t *pte = find_pte_pae_lpt(cr2);
	if (cr2 >= XEN_M2P_PAE)
	    return 0;
	if (*pte & _PAGE_USER)
	    return 0;
	/* is kernel page */
	*pte |= _PAGE_RW;
    }
    emu_faults[XEN_FAULT_PAGE_FAULT_FIX_RO]++;
    flush_tlb_addr(cr2);
    return 1;
}

/* --------------------------------------------------------------------- */
/* called from assembler                                                 */

asmlinkage void do_boot(struct regs_32 regs)
{
    uint32_t cr3_mfn;
    int vcpu = 0;
    
    printk(0, "this is emu32 (xenner %s)\n", VERSION);
    tss_init(vcpu);
    idt_init();
    paging_init();
    cr_init();
    pv_init(vcpu);
    printk(1, "emu setup done (tss, idt, paging, cr, pv)\n");

    /* guest paging setup */
    cr3_mfn = xen_ctxt.ctrlreg[3] >> PAGE_SHIFT;
    update_emu_mappings(cr3_mfn);
    write_cr3(frame_to_addr(cr3_mfn));
    
    /* guest register setup */
    regs.eax    = xen_ctxt.user_regs.eax;
    regs.ebx    = xen_ctxt.user_regs.ebx;
    regs.ecx    = xen_ctxt.user_regs.ecx;
    regs.edx    = xen_ctxt.user_regs.edx;
    regs.esi    = xen_ctxt.user_regs.esi;
    regs.edi    = xen_ctxt.user_regs.edi;
    regs.ebp    = xen_ctxt.user_regs.ebp;
    regs.eip    = xen_ctxt.user_regs.eip;
    regs.cs     = xen_ctxt.user_regs.cs;
    regs.eflags = xen_ctxt.user_regs.eflags;
    regs.esp    = xen_ctxt.user_regs.esp;
    regs.ss     = xen_ctxt.user_regs.ss;

    asm volatile("mov %0, %%ds;\n" :: "r" (xen_ctxt.user_regs.ds) : "memory");
    asm volatile("mov %0, %%es;\n" :: "r" (xen_ctxt.user_regs.es) : "memory");
    asm volatile("mov %0, %%fs;\n" :: "r" (xen_ctxt.user_regs.fs) : "memory");
    asm volatile("mov %0, %%gs;\n" :: "r" (xen_ctxt.user_regs.gs) : "memory");

    /* jump to kernel (via iret in asm code calling us) */
    printk(1, "booting kernel (entry %x:%x) ...\n", regs.cs, regs.eip);
}

asmlinkage void do_page_fault(struct regs_32 regs)
{
    uint32_t cr2 = read_cr2();

    emu_faults[XEN_FAULT_PAGE_FAULT]++;
    if (context_is_emu(&regs)) {
	if (fixup_extable(&regs))
	    return;
	print_page_fault_info(0, &regs, cr2);
	panic("ring0 (emu) page fault");
    }

    if (regs.error == 3) {
	/* kernel write to r/o page */
	if (pf_fixup_readonly(&regs, cr2))
	    return;
    }
    
    emu_faults[XEN_FAULT_PAGE_FAULT_GUEST]++;
    bounce_trap(&regs, 14, -1);
}

asmlinkage void do_event_callback(struct regs_32 regs)
{
    if (shared_info.vcpu_info[0].evtchn_upcall_mask) {
	/* FIXME */
	printk(2, "%s: IRQs off\n", __FUNCTION__);
	return;
    }
    emu_faults[XEN_FAULT_EVENT_CALLBACK]++;
    shared_info.vcpu_info[0].evtchn_upcall_mask = 1;
    bounce_trap(&regs, -1, CALLBACKTYPE_event);
}
